#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dlfcn.h>
#include </usr/include/mysql/mysql.h>
#include </usr/include/postgresql/libpq-fe.h> //in /usr/include
#include </usr/include/postgresql/libpq/libpq-fs.h>

#include "netmon_std.h"
#include "database_common.h"


/* network monitor - ed neville
 * Monitors the bytes of traffic on a network device
 * 2006-05-14 ed AT s5h DOT net - version 0.1 beta code
 * Use at your own risk. Author not responsible for use.
 *
 * Updates at http://www.s5h.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * $Id: database_common.c 61 2007-10-17 22:29:22Z ed $
 */

void free_database_struct( struct config *c ) {
	if( c->dbc.insert_space != NULL ) {
		free( c->dbc.insert_space );
	}
}

/**
 * this function just returns the database type specified in the first
 * argument
 */
int get_db_type( char *s ) {
	char types[4][10] = { "mysql", "sqlite", "pgsql", "\0" };
	int i;

	if( s == NULL ) {
		return(-1);
	}

	for( i = 0 ; types[i][0] != '\0' ; i++ ) {
		if( strncmp( types[i], s, strlen( types[i] ) ) == 0  ) {
			return(i);
		}
	}
	return(-1);
}

/**
 * this function just does the guts of the work required for
 * transferring the sql statement to the mysql library to carry out the
 * work
 */
int mysql_exec( struct config *c, char *sql_text ) {
	int retval = EXIT_SUCCESS;
	int (*mysql_query_ptr)(void *, void *);

	*(int **)(&mysql_query_ptr) = dlsym( c->mysql, "mysql_query" );

	if( mysql_query_ptr == NULL ) {
		fprintf( stderr, 
			"We cannot establish a handle for mysql_query" );
		return( EXIT_FAILURE );
	}

	retval = (int)(*mysql_query_ptr)( c->dbc.conn, sql_text );

	if( retval != 0 ) {
		retval = EXIT_FAILURE;
	}

	dlclose( mysql_query_ptr );

	return( retval );
}

/**
 * this function sets up the db handle so that other functions can make
 * calls to the db without having to set things up. pretty basic
 * really. this function is a bit longer than i would normally allow,
 * but that's because of all the dlsym work it's doing. it's
 * questionable if the handles should be freed when not used.
 */
int setup_mysql( struct config *c ) {
	int res;
	int (*mysql_conn_ptr)(void *);
	int (*mysql_real_connect_ptr)(void *, void *, void *, \
		void *, void *, int, void *, long );

	
	if( c->mysql == NULL ) {
		printf( "cannot start because of failure" );
		return( EXIT_FAILURE );
	}

	*(void **)(&mysql_conn_ptr) = dlsym( c->mysql, "mysql_init" );

	if( mysql_conn_ptr == NULL ) {
		/**
		 * we could not find the function
		 */
		return( EXIT_FAILURE );
	}

	c->dbc.conn = (int *)(*mysql_conn_ptr)( NULL );

	/**
	 * now we have a handle for the db. lets ehcke that it's not
	 * null shall we
	 */

	if( c->dbc.conn == NULL ) {
		return( EXIT_FAILURE );
	}

	*(void **)(&mysql_real_connect_ptr) = 
		dlsym( c->mysql, "mysql_real_connect" );

	if( mysql_real_connect_ptr == NULL ) {
		fprintf( stderr, "We cannot find a location for the mysql_real_connect function" );
		return( EXIT_FAILURE );
	}

	res = (int)(*mysql_real_connect_ptr)( (void *)c->dbc.conn,
		c->dbc.server,
		c->dbc.user, 
		c->dbc.password, 
		c->dbc.database,
		0,
		NULL,
		0 );
	
	if( !res ) {
		fprintf( stderr, 
			"Sorry there was an error attempting to "
			"establish a connection to the database. "
			"Perhaps the environment values are "
			"incorrect?\n" );
		return( EXIT_FAILURE );
	}

	if( c->dbc.insert_space == NULL ) {
		c->dbc.insert_space = (char *)malloc( BUFFER * 8 );
		if( c->dbc.insert_space == NULL ) {
			fprintf( stderr, 
				"We have run out of memory for the "
				"insert buffer. You system resources "
				"are extremely low" );
			return( EXIT_FAILURE );
		}
	}
	
	return( EXIT_SUCCESS );
}

int insert_mysql( struct config *c, struct interface_line *il ) {
	snprintf( c->dbc.insert_space, BUFFER*8, 
		"insert into netmon "
		"( netmon_interface, "
		"netmon_hostname, "
		"netmon_ibytes, netmon_obytes, "
		"netmon_ipackets, netmon_opackets ) "
		"values "
		"('%s', '%s', %lu, %lu, %lu, %lu )",
		il->name, c->dbc.hostname,
		il->ibytes, il->obytes,
		il->ipackets, il->opackets );

	return( mysql_exec( c, c->dbc.insert_space ) );
}



/**
 * these statements have been moved here so that we can setup the tables
 * as it should be a bit more straight forward from here
 */

/**
 * this function returns EXIT_FAILURE on error
 */
int sqlite_exec( struct config *c, char *statement ) {
	int (*sqlite3_exec)(void *,void *,void *, void *, void *);
	char *error;
	int res;
		
	*(int **)(&sqlite3_exec) = dlsym( c->sqlite, "sqlite3_exec" );

	if( sqlite3_exec == NULL ) {
		return( EXIT_FAILURE );
	}

	res = (int *)(*sqlite3_exec)( (void *)c->dbc.conn, 
		statement, NULL, 0, &error );
	if( res ) {
		return( EXIT_FAILURE );
	}

	dlclose( sqlite3_exec );
	
	return( EXIT_SUCCESS );
}

int setup_sqlite_create_tables( struct config *c ) {
	snprintf( c->dbc.insert_space, BUFFER*8,
		 "create table if not exists netmon( "
		 "netmon_id integer primary key autoincrement, "
		 "netmon_date timestamp default now, "
		 "netmon_interface varchar(255) not null default '', "
		 "netmon_hostname varchar(255) not null default '', "
		 "netmon_ibytes bigint  not null default 0, "
		 "netmon_obytes bigint not null default 0, "
		 "netmon_ipackets bigint not null default 0, "
		 "netmon_opackets bigint not null default 0 )" );

	if( sqlite_exec( c, c->dbc.insert_space ) == EXIT_FAILURE ) {
		return( EXIT_FAILURE );
	}

	snprintf( c->dbc.insert_space, BUFFER*8,
		"create index if not exists hostname_idx "
		"on netmon ( netmon_hostname )" );

	if( sqlite_exec( c, c->dbc.insert_space ) == EXIT_FAILURE ) {
		return( EXIT_FAILURE );
	}

	snprintf( c->dbc.insert_space, BUFFER*8,
		"create index if not exists interface_idx "
		"on netmon ( netmon_interface )" );
	
	if( sqlite_exec( c, c->dbc.insert_space ) == EXIT_FAILURE ) {
		return( EXIT_FAILURE );
	}

	return( EXIT_SUCCESS );
}

int setup_sqlite( struct config *c ) {
	int res;
	char *error;
	int (*fptr)(void*, void**);
	char (*msg)(void *);

	if( c->sqlite == NULL ) {
		printf( "cannot start before of failure\n" );
		return( EXIT_FAILURE );
	}

	/*fptr = (int (*)(int))dlsym( c->sqlite, "sqlite3_open" );*/
	*(void **)(&fptr) = dlsym( c->sqlite, "sqlite3_open" );

	if( fptr == NULL ) {
		return( EXIT_FAILURE );
	}

	/**
	 * call the function pointer with the sqlite paramter to obtain
	 * a db handle
	 */
	res = (*fptr)( c->dbc.database, (void *)(&c->dbc.conn) );

	if( res ) {
		*(void **)(&msg) = dlsym( c->sqlite, "sqlite3_errmsg" );
		if( fptr == NULL ) {
			return( EXIT_FAILURE );
		}

		error = (void *)(*msg)( &c->dbc.conn );
		
		fprintf( stderr, "Could not open: %s\n", error );
		return( EXIT_FAILURE );
	}

	c->dbc.insert_space = (char *)malloc( BUFFER*8 );
	if( c->dbc.insert_space == NULL ) {
		fprintf( stderr, "Not enough memory to allocate "
			"space for statement" );
		return( EXIT_FAILURE );
	}


	/**
	 * this function should not return to the caller the exit value,
	 * as it's misleading if the tables already exist.
	 */
	if( setup_sqlite_create_tables( c ) == EXIT_FAILURE ) {
		free( c->dbc.insert_space );
		/* return( EXIT_FAILURE ); */
	}


	return( EXIT_SUCCESS );
}

int insert_sqlite3( struct config *c, struct interface_line *il ) {
	snprintf( c->dbc.insert_space, BUFFER*8, 
		"insert into netmon "
		"( netmon_interface, "
		"netmon_hostname, "
		"netmon_ibytes, netmon_obytes, "
		"netmon_ipackets, netmon_opackets ) "
		"values "
		"('%s', '%s', %lu, %lu, %lu, %lu )",
		il->name, c->dbc.hostname,
		il->ibytes, il->obytes,
		il->ipackets, il->opackets );

	return( sqlite_exec( c, c->dbc.insert_space ) );
}

/**
 * this function puts the interface stats into an SQL statement for
 * execution
 */
int insert_pgsql( struct config *c, struct interface_line *il ) {
	snprintf( c->dbc.insert_space, BUFFER*8, 
		"insert into netmon "
		"( netmon_interface, "
		"netmon_hostname, "
		"netmon_ibytes, netmon_obytes, "
		"netmon_ipackets, netmon_opackets ) "
		"values "
		"('%s', '%s', %lu, %lu, %lu, %lu )",
		il->name, c->dbc.hostname,
		il->ibytes, il->obytes,
		il->ipackets, il->opackets );

	return( pgsql_exec( c, c->dbc.insert_space ) );
}

int pgsql_exec( struct config *c, char *statement ) {
	typedef PGresult *(*exec_func)();
	exec_func pqexec_func_ptr;
	PGresult *result;
	char *(*pqerrormessage_ptr)(void *);

	pqexec_func_ptr = (exec_func)dlsym( c->pgsql, "PQexec" );
	if( !pqexec_func_ptr ) { 
		/**
		 * TODO: is dlerror thread safe?
		 */
		fprintf( stderr, "We could not create a handle on "
			"PQexec: %s\n", dlerror() );
		return( EXIT_FAILURE );
	}

	result = (*pqexec_func_ptr)( c->dbc.conn, statement );
	dlclose( pqexec_func_ptr );

	if( result == NULL ) {
		*(void**)(&pqerrormessage_ptr) = dlsym( c->pgsql, 
			"PQerrorMessage" );

		if( pqerrormessage_ptr == NULL ) {
			fprintf( stderr, "cannot allocate handle: %s\n",
				dlerror() );
			return( EXIT_FAILURE );
		}

		fprintf( stderr, "We failed the execution: %s\n", 
			(char *)pqerrormessage_ptr( c->dbc.conn ) );
		return( EXIT_FAILURE );
	}

	return( EXIT_SUCCESS );
}

/**
 * this function just sets up the pointer for the executions. we have to
 * build the connection string up from the values pulled in from the
 * user environment - this is where snprintf really helps out as we can
 * just set a clause of buffer * 8.. if we cant fit it all in there then
 * the user has bigger problems. 
 */
int setup_pgsql( struct config *c ) {
	/* int *(ptr)(void *); */
	typedef PGconn *(*ConnFunc)();
	ConnFunc pqconnectdb_ptr;
	int retval = EXIT_SUCCESS;
	char conn_str[BUFFER*8];
	int (*pqstatus_ptr)(void *);
	int status;
	char *(*pqerrormessage_ptr)(void *);


	/**(void **)(&ptr) = dlsym( c->pgsql, "PQconnectdb" );*/

	pqconnectdb_ptr = (ConnFunc)dlsym( c->pgsql, "PQconnectdb" );
 	if( !pqconnectdb_ptr ) {
		fprintf( stderr, "We could not establish a handle "
			"for postgres connection\n" );
		return( EXIT_FAILURE );
	}

	snprintf( conn_str, BUFFER * 8,
		"host = '%s' "
		"user = '%s' "
		"dbname = '%s' "
		"password = '%s'",
		c->dbc.hostname, c->dbc.user, c->dbc.database, 
		c->dbc.password );

	c->dbc.conn = (*pqconnectdb_ptr)( conn_str );

	if( c->dbc.conn == NULL ) {
		fprintf( stderr, "We could not create a connection "
		"to the database.\n" );
		/**
		 * TODO: create function pointers for PQstatus and
		 * PQerrorMessage
		 */
		return( EXIT_FAILURE );
	}

	pqstatus_ptr = dlsym( c->pgsql, "PQstatus" );
	if( pqstatus_ptr == NULL ) {
		fprintf( stderr, "Cannot create handle for PQstatus: %s\n",
			dlerror() );
		return( EXIT_FAILURE );
	}

	status = pqstatus_ptr( c->dbc.conn );
	if( status != CONNECTION_OK ) {
		*(void**)(&pqerrormessage_ptr) = dlsym( c->pgsql, 
			"PQerrorMessage" );

		if( pqerrormessage_ptr == NULL ) {
			fprintf( stderr, "cannot allocate handle: %s\n",
				dlerror() );
			return( EXIT_FAILURE );
		}

		fprintf( stderr, "We failed the execution: %s\n", 
			(char *)pqerrormessage_ptr( c->dbc.conn ) );
		return( EXIT_FAILURE );
	}
	

	c->dbc.insert_space = (char *)malloc( BUFFER*8 );
	if( c->dbc.insert_space == NULL ) {
		fprintf( stderr, "Not enough memory to allocate "
			"space for statement" );
		return( EXIT_FAILURE );
	}

	return( retval );
}

int configure_db_settings_from_environemnt( struct config *c ) {
	char *user, *pass, *server, *database, *hostname;
	int ret = EXIT_SUCCESS;

	user = getenv( "NETMON_DB_USER" );
	pass = getenv( "NETMON_DB_PASS" );
	database = getenv( "NETMON_DB_DATABASE" );
	server = getenv( "NETMON_DB_SERVER" );
	hostname = getenv( "NETMON_HOSTNAME" );
	
	c->db_type = get_db_type( getenv( "NETMON_DB_TYPE" ) );

	if( user == NULL || pass == NULL 
		|| database == NULL  || server == NULL ) {
		return( EXIT_FAILURE );
	}

	strncpy( c->dbc.user, user, BUFFER );
	strncpy( c->dbc.password, pass, BUFFER );
	strncpy( c->dbc.database, database, BUFFER );
	strncpy( c->dbc.server, server, BUFFER );

	if( hostname == NULL ) {
		if( gethostname( c->dbc.hostname, BUFFER ) ) {
			/**
			 * cannot find hostname, we have a unnamed
			 * column
			 */
			strncpy( c->dbc.hostname, "Unknown host - try "
				"setting the environment "
				"NETMON_HOSTNAME", BUFFER );
		}
	}
	else {
		strncpy( c->dbc.hostname, hostname, BUFFER );
	}

	c->dbc.user[BUFFER-1] = '\0';
	c->dbc.password[BUFFER-1] = '\0';
	c->dbc.database[BUFFER-1] = '\0';
	c->dbc.server[BUFFER-1] = '\0';
	c->dbc.hostname[BUFFER-1] = '\0';

	switch( c->db_type ) {
		case mysql: {
			ret = setup_mysql(c);
			break;
		}
		case sqlite: {
			ret = setup_sqlite(c);
			break;
		}
		case pgsql: {
			ret = setup_pgsql(c);
			break;
		}
		default: {
			ret = EXIT_FAILURE;
			break;
		}
	}

	return( ret );
}

int is_db_environment_present( struct config *c ) {
	int ret = EXIT_FAILURE;
	if( getenv( "NETMONDB" ) ) {
		ret = configure_db_settings_from_environemnt( c );

	}
	return( ret );
}

int database_values( struct interface_line *il, struct config *c ) {

	if( is_db_environment_present( c ) == EXIT_FAILURE ) {
		fprintf( stderr, "No DB connection setup\n" );
		return( EXIT_FAILURE );
	}
	
	return( insert_db_row( c, il ) );
}

int insert_db_row( struct config *c, struct interface_line *il ) {
	int i;

	while( il != NULL ) {
		switch( c->db_type ) {
			case mysql: {
				i = insert_mysql( c, il );
				break;
			}
			case sqlite: {
				i = insert_sqlite3( c, il );
				break;
			}
			case pgsql: {
				i = insert_pgsql( c, il );
				break;
			}
		}
	
		if( i == EXIT_FAILURE ) {
			return( EXIT_FAILURE );
		}
		il = il->next;
	}
	return( EXIT_SUCCESS );
}


