#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dlfcn.h>

#include "netmon_std.h"

#ifdef DATABASE
#include "database_common.h"
#endif

/* network monitor - ed neville
 * Monitors the bytes of traffic on a network device
 *
 * this is just the "shell" file. if you wish to use the monitoring part
 * of this program to get interface values, look at netmon_linux.c and
 * netmon_bsd.c files which do the work.
 *
 * 2006-05-14 ed AT s5h DOT net - version 0.1 beta code
 * Use at your own risk. Author not responsible for use.
 *
 * Updates at http://www.s5h.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 * $Id: netmon_shell.c 62 2007-10-18 21:52:39Z ed $
 */


void print_list( struct interface_line *p, struct interface_line *pp, \
		struct interface_line *min );

void interface_cycle( struct config *c );

void list_interface_names(void);

void load_dynamic_libraries( struct config *c );

int main( int argc, char *argv[] ) {
	struct interface_line *il,*ll;
	int packets=0;
	struct config c;

	load_dynamic_libraries( &c );

	c.db_type = sqlite;

	if( argc == 1 ) {
		interface_cycle( &c );
		return( EXIT_SUCCESS );
	}
	
	il = getinterfaces();
	ll = il;

	if( argc > 1 && strncmp( argv[1], "--list", 6 ) == 0 ) {
		list_interface_names();
		exit( EXIT_SUCCESS );
	}

	#ifdef DATABASE
	if( argc > 1 && strncmp( argv[1], "--database", 10 ) == 0 ) {
		exit( database_values( il, &c ) );
	}
	#endif

	if( argc > 2 ) {
		if( argv[2][0] == 'p' ) {
			packets = 1;
		}
	}
	

	while( il != NULL ) {

		if( strncmp( il->name, argv[1], 20 ) == 0 ) {
			if( packets == 0 ) {
				printf( "%lu\n%lu\n", il->ibytes, \
						il->obytes );
			}
			else {
				printf( "%lu\n%lu\n", \
						il->ipackets, \
						il->opackets );
			}
			freeinterfaces(ll);
			exit( EXIT_SUCCESS );
		}
		il = il->next;
	}
		
	freeinterfaces(ll);	
	return( EXIT_FAILURE );
}

/**
 * 20070830 ewn many changes required to this function so that changes to 
 *              lists in the past do not effect the lists in the current 
 *              ;-) thanks to hondza for bringing this to my attention.
 */

void print_list( struct interface_line *p, \
	struct interface_line *last_second , \
	struct interface_line *last_minute ) {
	struct interface_line *pp, *min;

	while( p != NULL ) {
		if( pp == NULL ) {
			pp = p;
		}
		
		pp = if_name( p->name, last_second );
		min = if_name( p->name, last_minute );
		
		if( pp == NULL ) {
			pp = p;
		}

		if( min == NULL ) {
			min = p;
		}

		printf( 
			"%6s I:%5luMiB O:%5luMiB | "
			"I:%5luKiB O:%5luKiB | "
			"I:%5luKiB O:%5luKiB\n",

			p->name, p->ibytes/1024/1024, p->obytes/1024/1024,
			(p->ibytes-pp->ibytes)/1024, 
			(p->obytes-pp->obytes)/1024,
			(p->ibytes-min->ibytes)/1024,
			(p->obytes-min->obytes)/1024 );

		p=p->next;
		/* pp=pp->next;
		min=min->next; */
	}

	printf( "\n\n" );
}


void interface_cycle( struct config *c ) {
	struct interface_line *il, *ll = NULL, *min[60]/*,*temp*/;
	int count = 0, temp;
	
	#ifdef DATABASE
	int db_env = is_db_environment_present( c );
	#endif

	for( count = 0 ; count < 60 ; count++ ) {
		min[count] = getinterfaces();
	}
	count = 0;

	while( 1 ) {
	
		il = getinterfaces();
		freeinterfaces(min[count]);
		min[count] = getinterfaces();

		printf( "     [:: running total ::]   | "
			"  [:: reset/sec ::]   |    [:: reset/min ::]\n" );

		temp = count + 1;
		if( temp == 60 ) {
			temp = 0;

		}
		
		print_list( il, ll, min[temp] );
	
		#ifdef DATABASE
		if( db_env == EXIT_SUCCESS ) {
			insert_db_row( c, il );
		}
		#endif

		freeinterfaces(ll);
		if( ++count == 60 ) {
			count = 0;
		}
		ll = il;
		#ifndef TEST
		usleep( 1000000 );
		#endif
		fflush( stdout );
	}
}

void list_interface_names() {
	struct interface_line *p;

	p = getinterfaces();
	while( p != NULL ) {
		printf( "%s\n", p->name );
		p = p->next;
	}
	freeinterfaces(p);
}


/**
 * this procedure just loads some libraries into the config structure
 * for later use. if the libraries are not there then we just set the
 * config values to null and check that later when setting up the
 * configuration environment variables
 */
void load_dynamic_libraries( struct config *c ) {
	c->sqlite = dlopen( "libsqlite3.so", RTLD_NOW );
	if( dlerror() ) {
		c->sqlite = NULL;
	}

	c->mysql = dlopen( "libmysqlclient.so", RTLD_NOW );
	if( dlerror() ) {
		c->mysql = NULL;
	}

	c->pgsql = dlopen( "libpq.so", RTLD_NOW );
	if( dlerror() ) {
		c->pgsql = NULL;
	}
}

